:- comment(summary, "Interface to MySQL database").
:- comment(author, "Thomas H Kolbe, Universitaet Bonn").
:- comment(date, "$Date: 2003/04/24 16:41:49 $").
:- comment(desc, html("
    <H3>History</H3>
    This library has been written by
    <PRE>
	Thomas H. Kolbe, Univeritaet Bonn <tk@ikg.uni-bonn.de>
    </PRE>
    with modifications by
    <PRE>
	Horst Samulowitz, Aachen <horst@samulowitz.de>
    </PRE>
    The documentation was translated by
    <PRE>
	Joachim Schimpf, Imperial College London
    </PRE>
    
    <H3>Licence</H3>
    <P>
    This code may be freely copied and used.  It may be used in commercial
    applications, but cannot be sold for a profit.  If used within a
    commercial application, this code must be distributed for free along
    with the application.  The same applies for any work derived by
    modifying this code. 
    </P>
    <H3>General Hints</H3>
    <UL>
    <LI>Like in ECLiPSe, names (e.g. relation names) in MySQL are case
	sensitive.
    <LI>For all selection predicates, values of SQL type Char are returned
        as ECLiPSe atoms. If necessary, atom_string/2 can be used to convert
	them into ECLiPSe strings.
    <LI>For all selection predicates, null values of attributes are returned
        as the empty atom ''.
    <LI>External fact sets created via sql_external/2 do not honour schema
        modifications that are done while the database is connected.
	In particular, this can cause problems when attributes of the
	underlying relation are removed or renamed, or when the whole relation
	is removed or renamed. Since these actions can be performed from
	outside ECLiPSe, and since there is no notification mechanism,
	no automatic update is possible. 
    <LI>In addition, cached relations exhibit a version problem, since the
        cached facts are no longer up to date after database manipulations.
	In this case, the cached relation must either be recomputed by
	invoking sql_external(Name, cache) again, or the relation should
	be left uncached, using sql_external(Name, nocache).
    </UL>
    <H3>Error Messages</H3>
	<DL>
	<DT>type error in ...</DT><DD>
	This is an Eclipse error message indicating a type conflict.
	It can have several causes. Usually, some argument of an interface
	predicate has the wrong type.
	</DD>

	<DT>instantiation fault in ...</DT><DD>
	This is an Eclipse error message indicating a missing instantiation.
	It can have several causes. Usually, some argument of an interface
	predicate is uninstantiated, or not sufficiently instantiated.
	E.g. sql_select/2 requires a list skeleton of fixed length.
	</DD>

	<DT>Error: varlist too short!</DT><DD>
	The variable list and the attribute list in sql_query/5 or
	sql_querydistinct/5 have different lengths.
	</DD>

	<DT>Error: attributelist too short!</DT><DD>
	The variable list and the attribute list in sql_query/5 or
	sql_querydistinct/5 have different lengths.
	</DD>

	<DT>Error: attributelist must not be empty!</DT><DD>
	An empty attribute list was supplied, at least one is required.
	</DD>

	<DT>Error: varlist must not be empty!</DT><DD>
	An empty variable list was supplied, at least one element is required.
	</DD>

	<DT>Error: Cursor limit exhausted (max. no. is XX)!</DT><DD>
	The maximum number XX of open concurrently open cursors was exceeded.
	The first possibility to get around this problem is to rearrange the
	queries such that fewer open cursors are required, and make sure that
	resatisfiable selection predicates are cut or failed over.
	Another possible solution is to a priori retrieve all tuples into
	a list (using findall/3), and then process the list using member/2.
	After findall/3, the cursor is closed and can be reused.
	The maimum number of cursors can be increased by modifying the
	interface source (mysql.c) and recompile.
	</DD>

	<DT>Error: too many columns selected!</DT><DD>
	The maximum number of columns in a projection was exceeded in
        sql_query/5, sql_querydistinct/5 or sql_select/2.
	</DD>

	<DT>Error: select statement must not be empty!</DT><DD>
	An empty list was given to sql_select/2.
	</DD>

	<DT>Error: attributelist and variablelist must have the same length!</DT><DD>
	In sql_select/2, the number of projected attributes in the selection
	command must correspond to the length of the result variable list.
	</DD>

	<DT>Error: cursor# too big (range is 0-XX)!</DT><DD>
	This error should never occur and would indicate an internal
	problem in the interface code.
	</DD>

	<DT>MySQL-Fehler: ...</DT><DD>
	This is an error message from MySQL. Refer to the MySQL documentation.
	</DD>
	</DL>
    ")).

:- comment(sql_login/3, [
    summary:"Connect to a database",
    resat:no,
    fail_if:"Connection could not be established",
    amode:sql_login(+,+,+),
    args:["Database":"Atom or string",
	"Username":"Atom or string",
	"Password":"Atom or string"],
    desc:html("
    Connects the user Username to Database, which must be on the same
    host machine. The predicate fails if the connection could not be
    established. In this case, the reason can be determined via
    sql_lastmsg/3.
    "),
    eg:"
  ?- sql_login('my_database','testuser','geheim').
  yes.
    ",
    see_also:[sql_login/4,sql_logout/0]
]).

:- comment(sql_login/4, [
    summary:"Connect to a database",
    resat:no,
    fail_if:"Connection could not be established",
    amode:sql_login(+,+,+,+),
    args:["Database":"Atom or string",
	"Username":"Atom or string",
	"Password":"Atom or string",
	"Serverhost":"Atom or string"],
    desc:html("
    Like sql_login/3, but connects to Database on host Serverhost.
    The address of the database server can be given as host name
    (e.g. 'myserver.mydomain.com') or IP address (e.g. '192.168.0.10').
    "),
    eg:"
  ?- sql_login('my_database','testuser','geheim','192.168.0.10').
  yes.
    ",
    see_also:[sql_login/3,sql_logout/0]
]).

:- comment(sql_logout/0, [
    summary:"Disconnect from the database",
    resat:no,
    amode:sql_logout,
    desc:html("
    Disconnects from the database.
    "),
    eg:"
  ?- sql_login('my_database','testuser','geheim','192.168.0.10').
  yes.
    ",
    see_also:[sql_login/3,sql_login/4]
]).

:- comment(sql_insert/3, [
    summary:"Insert tuple into database",
    resat:no,
    fail_if:"Tuple cannot be inserted",
    amode:sql_insert(+,++,++),
    args:["Table":"Atom or string",
	"Attributelist":"List of atoms or strings",
	"Valuelist":"List of ground terms"],
    desc:html("
    Inserts a tuple into the relation Table. The tuple is specified by
    Attributelist, which is the list of attribute names, and Valuelist,
    which is the list of corresponding values to be inserted.
    Both lists must be of the same length.
    "),
    eg:"
  % insert a person into relation 'Adressen'
  ?- sql_insert('Adressen',['Id','Vorname','Name'],
		  [113,'Karl','Krause']).
		    yes.
    ",
    see_also:[sql_delete/5]
]).


:- comment(sql_delete/5, [
    summary:"Delete tuple(s) from database",
    resat:no,
    amode:sql_delete(+,+,++,++,-),
    args:["Table":"Atom or string",
	"Where":"Atom or string",
	"Attributelist":"List of atoms or strings",
	"Valuelist":"List of ground terms",
	"RowsDeleted":"Variable or integer"],
    desc:html("
    Used to delete tuples from the relation Table. The set of tuples to
    be deleted can be restricted using both Where and Attributelist/Valuelist.
    The string Where will be inserted literally after the keyword WHERE
    in the SQL DELETE instruction. If Attributelist and Valuelist are not
    empty, then only tuples whose attributes in Attributelist match the
    corresponding values in Valuelist, will be deleted.
    Both lists must have the same length. After deletion, RowsDeleted
    will be unified with the number of deleted rows. Even if the unification
    fails, the rows will remain deleted.
    "),
    eg:"
    % delete all 5 tuples of relation 'Adressen'
    ?- sql_delete('Adressen','',[],[],R).
    R = 5
    yes.

    % Delete from relation 'Adressen' (which has the attributes Id and Name)
    % all people with name \"Schmidt\" whose Id is also stored in the relation
    % 'Loeschbar'
    ?- sql_delete(\"Adressen\",\"Id in (select Id from Loeschbar)\",
                [\"Name\"],[\"Schmidt\"],_).
    yes.
    ",
    see_also:[sql_insert/3]
]).


:- comment(sql_exec/4, [
    summary:"Execute an SQL statement",
    resat:no,
    amode:sql_exec(+,-,-,-),
    args:["Statement":"Atom or string",
	"ReturnCode":"Variable or number",
	"ReturnString":"Variable or string",
	"RowsProcessed":"Variable or integer"],
    desc:html("
    Executes the given SQL Statement, and unifies ReturnCode with MySQL's
    return code and ReturnString with the corresponding message.
    If appropriate, RowsProcessed will be unified with the number of rows
    processed, otherwise with 0.
    The effect of the statement execution will persist regardless of
    whether the unification of ReturnCode, ReturnString or RowsProcessed
    succeeds or fails.
    "),
    eg:"
    % create a new relation 'Adressen2'
    ?- sql_exec(\"create table Adressen2 (Id integer(8), Name char(30))\",
    ReturnCode,ReturnMsg,_).
    ReturnCode = 0
    ReturnMsg = \"\"
    yes.

    % Tranfer all Names and Ids from 'Adressen' to 'Adressen2',
    % assuming that 'Adressen' contains 12 tuples
    ?- sql_exec(\"insert into Adressen2 select Id,Name from Adressen\",
	      ReturnCode,_,Rows).
    ReturnCode = 0
    Rows = 12
    yes.
    ",
    see_also:[sql_insert/3]
]).


:- comment(sql_query/5, [
    summary:"Query the database",
    resat:yes,
    amode:sql_query(++,++,+,+,?),
    args:["Attributelist":"List of atoms or strings",
	"Tables":"Atom or string",
	"Where":"Atom or string",
	"MoreOptions":"Atom or string",
	"Valuelist":"Variable or partially instantiated list"],
    desc:html("<P>
    Used to select and retrieve tuples from one or more database relations. 
    The tuples are retrieved one at a time on backtracking, i.e. every
    attempt at resatisfying the predicate will give another matching tuple.
    If there are no further tuples, the predicate fails.
    </P><P>
    Attributelist specifies the attributes of the selection to be projected.
    Theses attributes of the retrieved tuples will be unified with the
    corresponding elements of Valuelist.
    </P><P>
    Tables is the name of a relation, or several names separated by commas
    within the string (Tables is inserted after the FROM keyword in the SQL
    statement).
    </P><P>
    Where is the selection condition im SQL syntax and will be inserted after
    the WHERE keyword in the SQL statement.
    </P><P>
    MoreOptions can be used to specify further SQL clauses, e.g. GROUP BY
    or ORDER BY, as well as for combining selections via UNION, INTERSECT
    etc. (MoreOptions will be appended to the SQL statement constructed
    so far, separated by a blank space).
    </P><P>
    Important note: For every new invocation of sql_query/5 the interface
    opens a new cursor into the relation resulting from the selection.
    This cursor only gets closed when either all tuples have been read,
    or when backtracking retrieval was terminated prematurely by a cut (!/0).
    In complex programs this can exceed the maximum number of cursors
    allowed by the interface. A possible solution is to a priori retrieve all
    tuples into a list (using findall/3), and then process the list
    using member/2.
    </P>
    "),
    eg:"
  % List the tuples of relation Adressen (which has attributes
  % Id Vorname and Name), assuming there are two such tuples:

  ?- sql_query(['Id','Vorname','Name'],'Adressen','','',[I,V,N]).
  I = 113
  V = 'Karl'
  N = 'Krause'  More? (;)

  I = 17
  V = 'Berta'
  N = 'Braesig' More? (;)

  no (more) solution.



  % Find the names of persons, and their department (Abteilung),
  % using a join of relations Adressen and Arbeitsplatz.
  % The join is made via the attribute Id which is present in
  % both relations

  ?- sql_query(['Name','Abteilung'],'Adressen,Arbeitsplatz',
                'Adressen.Id=Arbeitsplatz.Id','',[Name,Abteilung]).
  Name = 'Braesig'
  Abteilung = 'Buchhaltung'   More? (;)

  Name = 'Krause'
  Abteilung = 'Logistik'      More? (;)

  no (more) solution.



  % List the Id numbers of persond which occur both in relation
  % Addressen and Arbeitsplatz. It is implemeted here using a set
  % intersection, in order to show a use of the MoreOptions argument.
  % Normally, this problem would better be solved using a join.

  ?- sql_query(['Id'],'Adressen','',
                'intersect select Id from Arbeitsplatz',[I]).
  I = 113      More? (;)

  I = 17       More? (;)

  no (more) solution.
    ",
    see_also:[sql_insert/3,sql_querydistinct/5,sql_select/2]
]).

:- comment(sql_querydistinct/5, [
    summary:"Query the database without duplicates",
    resat:yes,
    amode:sql_querydistinct(++,++,+,+,?),
    args:["Attributelist":"List of atoms or strings",
	"Tables":"Atom or string",
	"Where":"Atom or string",
	"MoreOptions":"Atom or string",
	"Valuelist":"Variable or partially instantiated list"],
    desc:html("<P>
    This is exactly like sql_query/5, except that duplicates are eliminated.
    </P><P>
    In practice this means that the SQL keyword DISTINCT is inserted into
    the corresponding SQL statement.
    </P>
    "),
    see_also:[sql_insert/3,sql_query/5,sql_select/2]
]).

:- comment(sql_select/2, [
    summary:"Query the database using SQL SELECT statement",
    resat:yes,
    amode:sql_select(++,+),
    args:["StatementList":"List of atoms, strings and numbers",
	"Valuelist":"Partially instantiated, proper list"],
    desc:html("<P>
    This is the generic variant of the selection command. StatementList
    is a fully instantiated list of atoms, strings and numbers, which
    get concatenated, separated by blanks, and prefixed with SELECT,
    to form a valid SQL SELECT statement. 
    </P><P>
    Valuelist must be a list of fixed length, containing variables or
    values. For every result tuple, the tuple gets unified with Valuelist,
    such that the first list element corresponds to the first attribute of
    the selection, the second to the second, etc. Every resatisfaction
    of the predicate delivers another tuple of the selection result.
    </P><P>
    The important difference between sql_select/2 and sql_query/5 is that
    sql_select/2 is more general, but does not take advantage of elements
    in Valuelist which are already instantiated: first, all tuples of the
    selection are retrieved, and only then do they get unified with
    Valuelist, which may lead to failure. With sql_query/5, instantiated
    elements of Valuelist will be incorporated into the selection
    criteria of the SELECT statement. As a consequence, the result set
    is already (possibly vastly) reduced on the database side.
    </P><P>
    Hint: In SQL, strings must be enclosed in single quotes. If such strings
    are to be used within Statementlist, it is convenient to surround the
    single-quoted SQL strings with double quotes (which can contain single
    quotes without the need for further escape symbols).
    </P><P>
    Important note: For every new invocation of sql_query/5 the interface
    opens a new cursor into the relation resulting from the selection.
    This cursor only gets closed when either all tuples have been read,
    or when backtracking retrieval was terminated prematurely by a cut (!/0).
    In complex programs this can exceed the maximum number of cursors
    allowed by the interface. A possible solution is to a priori retrieve all
    tuples into a list (using findall/3), and then process the list
    using member/2.
    </P>
    "),
    eg:"
    % The relation Adressen has attributes Id, Vorname and Name.
    % Find Vorname and Name of persons in relation Adressen, whose
    % Id lie in the interval (Min,Max).
    % Assuming there are two solution in the interval (17,200):

      ?- Min=17, Max=200,
	 sql_select(['Vorname, Name from Adressen where Id between',
		     Min,and,Max],[V,N]).
      V = 'Karl'
      N = 'Krause'  More? (;)

      V = 'Berta'
      N = 'Braesig' More? (;)

      no (more) solution.
    ",
    see_also:[sql_query/5,sql_querydistinct/5,sql_insert/3]
]).

:- comment(sql_lastmsg/3, [
    summary:"Get the status of the last MySQL operation",
    amode:sql_lastmsg(?,?,?),
    args:["ReturnCode":"Variable or integer",
	"ReturnString":"Variable or string",
	"RowsProcessed":"Variable or integer"],
    desc:html("<P>
    This predicate unifies the three paramenters with the status of the last
    operation performed by MySQL. Please refer to the MySQL documentation
    for the meaning of the return codes. If the last operation selected or
    modified any tuples, RowsProcessed indocates how many (0 otherwise).
    </P>
    "),
    see_also:[]
]).

:- comment(sql_commit/0, [
    summary:"Commit transaction (not implemented)",
    fail_if:none,
    desc:html("<P>
    This predicate has no function in MySQL and exists for compatibility with
    the Eclipse-Oracle interface. MySQL does not currently support the
    transaction concept, all database operations get executed immediately
    and finally. The predicate always succeeds.
    </P>
    "),
    see_also:[sql_rollback/0]
]).

:- comment(sql_rollback/0, [
    summary:"Roll back a transaction (not implemented)",
    fail_if:none,
    desc:html("<P>
    This predicate has no function in MySQL and exists for compatibility with
    the Eclipse-Oracle interface. MySQL does not currently support the
    transaction concept, all database operations get executed immediately
    and finally. Since database manipulations cannot be rolled back in MySQL,
    the predicate prints a warning, but succeeds nevertheless.
    </P>
    "),
    see_also:[sql_commit/0]
]).

:- comment(sql_external/2, [
    summary:"Make a database relation available as a predicate",
    amode:sql_external(++,++),
    args:["Table":"Atom",
    	"CacheMode":"Atom"],
    desc:html("<P>
    This predicate makes a data base relation available as a predicate,
    (similar to a set of facts). This is done by reading from the database
    the definition of relation Table, and generating a predicate whose
    name is the concatenation of 'r_' and the relation name Table.
    The arity and the order of arguments correspond to the definition of
    the relation's attributes.
    </P><P>
    CacheMode is the atom 'cache' or 'nocache'. It specifies whether the
    contents of the entity should be cached in main memory. Caching means
    that the whole relation gets retrieved from the database, every tuple
    converted to a fact, and the result compiled as a predicate. For large
    relations, this can take time and consume large amounts of memory.
    </P><P>
    Important note: For uncached predicates, every new invocation
    opens a new cursor into the relation resulting from the selection.
    This cursor only gets closed when either all tuples have been read,
    or when backtracking retrieval was terminated prematurely by a cut (!/0).
    In complex programs this can exceed the maximum number of cursors
    allowed by the interface. A possible solution is to a priori retrieve all
    tuples into a list (using findall/3), and then process the list
    using member/2.
    </P>
    "),
    eg:"
    % Make the relation Adressen (which has attributes Id, Name, Vorname)
    % available as an (uncached) predicate
      ?- sql_external('Adressen',nocache).
      yes.

      ?- r_Adressen(I,N,V).
      I = 113
      N = 'Krause'
      V = 'Karl'     More? (;)

      I = 17
      N = 'Braesig'
      V = 'Berta'
      yes.
    ",
    see_also:[]
]).

:- comment(sql_showdb/0, [
    summary:"Display all relations defined by the currently logged in user",
    desc:html("<P>
    Display all relations defined by the currently logged in user.
    This will always succeed as long asn the database is connected and
    no database error occurs.
    </P>
    "),
    see_also:[]
]).

:- comment(sql_desc/1, [
    summary:"Display the schema of the given relation",
    amode:sql_desc(++),
    args:["Table":"Atom"],
    desc:html("<P>
    Display the schema of the given relation. The output is similar to
    the result of the SQL command DESCRIBE.
    This will always succeed as long asn the database is connected,
    no database error occurs, and the relation Table exists.
    </P>
    "),
    see_also:[]
]).



